from curia.modeling.metrics import MetricsInterface
from curia.session import Session

from typing import Dict, List
import pandas as pd

class TrainedModelJob:
    def __init__(self, session: Session, model_job_id: str) -> None:
        self.session = session
        self.model_job_id = model_job_id

    def get_basic_metrics(self) -> Dict[str, float]:
        """Extract basic metrics from a model job.

        Returns:
            Dict[str, float]: A dictionary of the metrics.
        """
        basic_metrics = MetricsInterface.get_basic_metrics(self.session, self.model_job_id)
        return basic_metrics

    def all_named_metrics(self) -> List[str]:
        """Get the possible named metrics

        Returns:
            List[str]: The list of named metrics.
        """
        metric_names = MetricsInterface.get_registered_metrics()
        return metric_names

    def get_metrics(self, metrics_to_get: List[str]) -> Dict[str, float or pd.DataFrame]:
        """Gets a set of metrics listed in 'metrics' from the model trained or model_job_id.

        Args:
            metrics_to_get (List[str]): The list of metrics to get. A full list can be found with the `all_named_metrics` method.

        Returns:
            Dict[str, float or pd.DataFrame]: A dictionary of the metrics.
        """
        output_metrics_dict = MetricsInterface.get_metrics(self.session, metrics_to_get, self.model_job_id)
        return output_metrics_dict



# coding: utf-8

"""
    Curia Platform API

    These are the docs for the curia platform API. To test, generate an authorization token first.  # noqa: E501

    OpenAPI spec version: 3.10.4
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class CreateOrganizationSettingDto(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'models': 'object',
        'min_claim_date': 'datetime',
        'max_claim_date': 'datetime',
        'max_cohort_windows': 'float',
        'pipeline_version': 'str',
        'risk_version': 'str',
        'impactability_version': 'str',
        'primary_database': 'str',
        'features_database': 'str'
    }

    attribute_map = {
        'models': 'models',
        'min_claim_date': 'minClaimDate',
        'max_claim_date': 'maxClaimDate',
        'max_cohort_windows': 'maxCohortWindows',
        'pipeline_version': 'pipelineVersion',
        'risk_version': 'riskVersion',
        'impactability_version': 'impactabilityVersion',
        'primary_database': 'primaryDatabase',
        'features_database': 'featuresDatabase'
    }

    def __init__(self, models=None, min_claim_date=None, max_claim_date=None, max_cohort_windows=None, pipeline_version=None, risk_version=None, impactability_version=None, primary_database=None, features_database=None):  # noqa: E501
        """CreateOrganizationSettingDto - a model defined in Swagger"""  # noqa: E501
        self._models = None
        self._min_claim_date = None
        self._max_claim_date = None
        self._max_cohort_windows = None
        self._pipeline_version = None
        self._risk_version = None
        self._impactability_version = None
        self._primary_database = None
        self._features_database = None
        self.discriminator = None
        if models is not None:
            self.models = models
        if min_claim_date is not None:
            self.min_claim_date = min_claim_date
        if max_claim_date is not None:
            self.max_claim_date = max_claim_date
        self.max_cohort_windows = max_cohort_windows
        if pipeline_version is not None:
            self.pipeline_version = pipeline_version
        if risk_version is not None:
            self.risk_version = risk_version
        if impactability_version is not None:
            self.impactability_version = impactability_version
        if primary_database is not None:
            self.primary_database = primary_database
        if features_database is not None:
            self.features_database = features_database

    @property
    def models(self):
        """Gets the models of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The models of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: object
        """
        return self._models

    @models.setter
    def models(self, models):
        """Sets the models of this CreateOrganizationSettingDto.


        :param models: The models of this CreateOrganizationSettingDto.  # noqa: E501
        :type: object
        """

        self._models = models

    @property
    def min_claim_date(self):
        """Gets the min_claim_date of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The min_claim_date of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: datetime
        """
        return self._min_claim_date

    @min_claim_date.setter
    def min_claim_date(self, min_claim_date):
        """Sets the min_claim_date of this CreateOrganizationSettingDto.


        :param min_claim_date: The min_claim_date of this CreateOrganizationSettingDto.  # noqa: E501
        :type: datetime
        """

        self._min_claim_date = min_claim_date

    @property
    def max_claim_date(self):
        """Gets the max_claim_date of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The max_claim_date of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: datetime
        """
        return self._max_claim_date

    @max_claim_date.setter
    def max_claim_date(self, max_claim_date):
        """Sets the max_claim_date of this CreateOrganizationSettingDto.


        :param max_claim_date: The max_claim_date of this CreateOrganizationSettingDto.  # noqa: E501
        :type: datetime
        """

        self._max_claim_date = max_claim_date

    @property
    def max_cohort_windows(self):
        """Gets the max_cohort_windows of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The max_cohort_windows of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: float
        """
        return self._max_cohort_windows

    @max_cohort_windows.setter
    def max_cohort_windows(self, max_cohort_windows):
        """Sets the max_cohort_windows of this CreateOrganizationSettingDto.


        :param max_cohort_windows: The max_cohort_windows of this CreateOrganizationSettingDto.  # noqa: E501
        :type: float
        """
        if max_cohort_windows is None:
            raise ValueError("Invalid value for `max_cohort_windows`, must not be `None`")  # noqa: E501

        self._max_cohort_windows = max_cohort_windows

    @property
    def pipeline_version(self):
        """Gets the pipeline_version of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The pipeline_version of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: str
        """
        return self._pipeline_version

    @pipeline_version.setter
    def pipeline_version(self, pipeline_version):
        """Sets the pipeline_version of this CreateOrganizationSettingDto.


        :param pipeline_version: The pipeline_version of this CreateOrganizationSettingDto.  # noqa: E501
        :type: str
        """

        self._pipeline_version = pipeline_version

    @property
    def risk_version(self):
        """Gets the risk_version of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The risk_version of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: str
        """
        return self._risk_version

    @risk_version.setter
    def risk_version(self, risk_version):
        """Sets the risk_version of this CreateOrganizationSettingDto.


        :param risk_version: The risk_version of this CreateOrganizationSettingDto.  # noqa: E501
        :type: str
        """

        self._risk_version = risk_version

    @property
    def impactability_version(self):
        """Gets the impactability_version of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The impactability_version of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: str
        """
        return self._impactability_version

    @impactability_version.setter
    def impactability_version(self, impactability_version):
        """Sets the impactability_version of this CreateOrganizationSettingDto.


        :param impactability_version: The impactability_version of this CreateOrganizationSettingDto.  # noqa: E501
        :type: str
        """

        self._impactability_version = impactability_version

    @property
    def primary_database(self):
        """Gets the primary_database of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The primary_database of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: str
        """
        return self._primary_database

    @primary_database.setter
    def primary_database(self, primary_database):
        """Sets the primary_database of this CreateOrganizationSettingDto.


        :param primary_database: The primary_database of this CreateOrganizationSettingDto.  # noqa: E501
        :type: str
        """

        self._primary_database = primary_database

    @property
    def features_database(self):
        """Gets the features_database of this CreateOrganizationSettingDto.  # noqa: E501


        :return: The features_database of this CreateOrganizationSettingDto.  # noqa: E501
        :rtype: str
        """
        return self._features_database

    @features_database.setter
    def features_database(self, features_database):
        """Sets the features_database of this CreateOrganizationSettingDto.


        :param features_database: The features_database of this CreateOrganizationSettingDto.  # noqa: E501
        :type: str
        """

        self._features_database = features_database

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(CreateOrganizationSettingDto, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateOrganizationSettingDto):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



import click

from curia.infra.container.deploy import build_container_tasks, deploy_container_tasks
from curia.infra.container.task import ContainerTaskDefinition
from curia.infra.types import ContainerTaskDeployResourceNamingConfig
from curia.infra.utils import load_tasks, describe_tasks


@click.group()
def container():
    pass


@container.command()
@click.argument("project_dir")
@click.argument("package")
@click.option("--verbose", is_flag=True, default=False, help="Toggle verbose output")
def describe(project_dir, package, verbose) -> None:
    """
    Describes all tasks
    """
    tasks = load_tasks(project_dir, package, task_type=ContainerTaskDefinition)
    if len(tasks) == 0:
        click.echo("No tasks found")
        return
    describe_tasks(tasks, verbose)


@container.command()
@click.argument("project_dir")
@click.argument("package")
@click.argument("prefix")
@click.option("--version", default="latest", help="The version to build")
@click.option("--cache-from-ecr", is_flag=True, default=False, help="Whether or not to use the ECR cache")
@click.option("--build-arg", multiple=True, help="Build arguments to pass to the docker build command")
def build(project_dir, package, prefix, version, cache_from_ecr, build_arg) -> None:
    """
    Builds all tasks
    """
    tasks = load_tasks(project_dir, package, task_type=ContainerTaskDefinition)
    if len(tasks) == 0:
        click.echo("No tasks found")
        return
    build_container_tasks(
        project_dir,
        tasks,
        resource_naming_config=ContainerTaskDeployResourceNamingConfig(
            image_prefix=prefix,
            platform_task_prefix=prefix,
            platform_workflow_prefix=prefix,
        ),
        version=version,
        cache_from_ecr=cache_from_ecr,
        build_args=build_arg,
    )


@container.command()
@click.argument("project_dir")
@click.argument("package")
@click.argument("env")
@click.argument("prefix")
@click.argument("batch_job_role")
@click.argument("ecs_task_role")
@click.argument("ecs_execution_role")
@click.option("--version", default="latest", help="The version to deploy")
@click.option("--do-build", is_flag=True,
              help="Whether or not to build the tasks before deploying")
@click.option("--build-arg", multiple=True, help="Build arguments to pass to the docker build command")
def deploy(
        project_dir,
        package,
        env,
        prefix,
        batch_job_role,
        ecs_task_role,
        ecs_execution_role,
        version,
        do_build,
        build_arg
) -> None:
    """
    Deploys all tasks
    """
    tasks = load_tasks(project_dir, package, task_type=ContainerTaskDefinition)
    if len(tasks) == 0:
        click.echo("No tasks found")
        return
    click.echo('Loaded tasks:')
    for task in tasks:
        click.echo(f'    {task.name} ({task.task_slug})')
    if do_build:
        build_container_tasks(
            project_dir,
            tasks,
            resource_naming_config=ContainerTaskDeployResourceNamingConfig(
                image_prefix=prefix,
                platform_task_prefix=prefix,
                platform_workflow_prefix=prefix,
            ),
            version=version,
            cache_from_ecr=True,
            build_args=build_arg,
        )
    deploy_container_tasks(
        tasks,
        version=version,
        env=env,
        batch_job_role=batch_job_role,
        ecs_task_role=ecs_task_role,
        ecs_execution_role=ecs_execution_role,
        resource_naming_config=ContainerTaskDeployResourceNamingConfig(
            image_prefix=prefix,
            platform_task_prefix=prefix,
            platform_workflow_prefix=prefix,
        ),
    )


if __name__ == "__main__":
    container()



try:
    import docker
except ImportError as exc:
    raise ImportError(
        "Some optional dependencies required for curia.infra.container are missing. "
        "Please install them via `poetry install --with=container_infra` if developing or"
        " `pip install curia[container_infra]` if using curia as a library."
    ) from exc



"""
This script deploys container tasks defined in airml_analytics
"""
import base64
import json
import os

import click
import databricks.sdk.service.compute as databricks_compute
import databricks.sdk.service.jobs as databricks_jobs
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.workspace import ExportFormat
from typing import List

from curia.infra.common import update_curia_task_and_stub_workflow
from curia.infra.databricks.task import DatabricksTaskDefinition
from curia.infra.types import DatabricksTaskDeployResourceNamingConfig, DatabricksTaskCodeSourceType


def deploy_databricks_tasks(
        tasks: List[DatabricksTaskDefinition],
        env: str,
        resource_naming_config: DatabricksTaskDeployResourceNamingConfig
) -> None:
    """
    Deploys all tasks as databricks jobs
    :param tasks: List of tasks to deploy
    :param env: The environment to deploy to
    :param resource_naming_config: The resource naming config to use
    :return:
    """
    db_client = WorkspaceClient(
        host=os.environ["DATABRICKS_HOST"],
        token=os.environ["DATABRICKS_TOKEN"]
    )
    for task in tasks:
        databricks_job_id = create_or_update_databricks_job(
            task,
            env,
            db_client,
            resource_naming_config=resource_naming_config
        )
        update_curia_task_and_stub_workflow(
            task,
            env,
            resource_naming_config=resource_naming_config,
            task_specific_context={
                "databricks_job_id": databricks_job_id
            }
        )
        click.echo(f"Deployed {task.task_slug} to {env}")


def create_or_update_databricks_job(
        task: DatabricksTaskDefinition,
        env: str,
        db_client: WorkspaceClient,
        resource_naming_config: DatabricksTaskDeployResourceNamingConfig):
    """
    Creates or updates a databricks job in the Databricks workspace, configuring it to run the task with access to
    appropriate S3 resources and the correct cluster configuration.
    """
    databricks_task_name = f"{resource_naming_config.databricks_workflow_prefix}{env}-{task.task_slug}"
    notebook_path = _create_databricks_notebook(
        task,
        env,
        db_client,
        databricks_task_name
    )
    kwargs = dict(
        name=databricks_task_name,
        tasks=[
            databricks_jobs.Task(
                description="",
                existing_cluster_id="",
                notebook_task=databricks_jobs.NotebookTask(
                    base_parameters={
                        "GENERICS_ONLY": "false",
                        "code_version": "NA",
                        "data_version": "NA",
                        "env": "PROD",
                    },
                    notebook_path=notebook_path,
                    source=databricks_jobs.Source.WORKSPACE
                ),
                task_key="job",
                new_cluster=databricks_compute.ClusterSpec(
                    autoscale=databricks_compute.AutoScale(
                        max_workers=task.max_workers,
                        min_workers=task.min_workers
                    ),
                    aws_attributes=databricks_compute.AwsAttributes(
                        first_on_demand=1,
                        availability=databricks_compute.AwsAvailability.SPOT_WITH_FALLBACK,
                        zone_id="us-east-1c",
                        instance_profile_arn="arn:aws:iam::452233835093:instance-profile/Databricks-S3-prod",
                        spot_bid_price_percent=100,
                        ebs_volume_count=0
                    ),
                    spark_version="9.1.x-scala2.12",
                    spark_conf={
                        "spark.hadoop.fs.s3a.acl.default": "BucketOwnerFullControl",
                        "spark.sql.shuffle.partitions": "200",
                        "spark.sql.legacy.timeParserPolicy": "LEGACY",
                        "spark.sql.legacy.parquet.datetimeRebaseModeInWrite": "LEGACY"
                    },
                    node_type_id="i3.2xlarge",
                    driver_node_type_id="i3.4xlarge",
                    ssh_public_keys=[],
                    custom_tags={},
                    spark_env_vars={},
                    enable_elastic_disk=True,
                    cluster_source=databricks_compute.ClusterSource.JOB,
                    enable_local_disk_encryption=False,
                    runtime_engine=databricks_compute.RuntimeEngine.STANDARD
                )
            )
        ],
    )
    # check if job exists
    jobs = db_client.jobs.list(name=databricks_task_name)
    jobs_list = list(jobs)
    job = None
    if len(jobs_list) > 1:
        for job in jobs_list[:-1]:
            click.echo(f"Deleting excess job {job.job_id}")
            db_client.jobs.delete(job_id=job.job_id)
        job = jobs_list[-1]
    elif len(jobs_list) == 1:
        job = jobs_list[0]

    if job is not None:
        click.echo(f"Job {databricks_task_name} exists, updating")
        db_client.jobs.reset(job_id=job.job_id, new_settings=databricks_jobs.JobSettings(**kwargs))

    else:
        click.echo(f"Job {databricks_task_name} doesn't exist, creating")
        job = db_client.jobs.create(**kwargs)
    return job.job_id


def _build_cell(code: str):
    """
    Builds a databricks notebook cell from a string of code
    """
    return {
        "cell_type": "code",
        "execution_count": None,
        "metadata": {},
        "source": [line + "\n" for line in code.split("\n")],
        "outputs": []
    }


def _build_notebook(cells: List[dict]):
    """
    Builds a databricks notebook from a list of cells
    """
    return {
        "cells": cells,
        "metadata": {},
        "nbformat": 4,
        "nbformat_minor": 0
    }


def _build_task_head_notebook(env: str, task: DatabricksTaskDefinition, curia_api_endpoint: str):
    """
    Creates a notebook for the task head, which is responsible for setup surrounding the code execution and
    calling the task code.
    """
    cells = []

    package = task.code_src_cfg["package"]
    if "package" not in task.code_src_cfg:
        raise ValueError("package not specified in code_src_cfg")

    install_cell = _build_install_cell(task, package)

    cells.append(install_cell)
    cells.append(_build_cell(f"""
import {package}

print({package}.__version__)
"""))

    cells.append(_build_cell(f"""
from {task.module} import {task.function_name}
""".strip()))

    cells.append(_build_cell(f"""
environment = dbutils.widgets.get("env")
execution_id = dbutils.widgets.get("executionId")
api_token = dbutils.secrets.get("airml-etl", "curia-{env}-api-token")
api_endpoint = "{curia_api_endpoint}"
""".strip()))

    cells.append(_build_cell(f"""
{task.function_name}.run(
    task_execution_id=execution_id,
    api_token=api_token
    api_endpoint=api_endpoint,
    spark=spark,
    dbutils=dbutils
)
""".strip()))
    return _build_notebook(cells)


def _build_install_cell(task, package):
    if task.code_src_type == DatabricksTaskCodeSourceType.S3WHEEL:
        return _build_install_cell_s3wheel(task, package)
    if task.code_src_type == DatabricksTaskCodeSourceType.PRIVATE_PYPI:
        return _build_install_cell_private_pipy(package)
    raise ValueError(f"Unsupported code_src_type {task.code_src_type}")


def _build_install_cell_s3wheel(task, package):
    if "bucket" not in task.code_src_cfg:
        raise ValueError("bucket not specified in code_src_cfg")
    bucket = task.code_src_cfg["bucket"]
    install_cell = _build_cell(f"""
code_version = dbutils.widgets.get("code_version")
import boto3

client = boto3.client("s3")

client.download_file("{bucket}", f"{package}-{{code_version}}-py3-none-any.whl", f"/dbfs/{package}-{{code_version}}-py3-none-any.whl")

%pip install /dbfs/{package}-$code_version-py3-none-any.whl
    """.strip())
    return install_cell


def _build_install_cell_private_pipy(package):
    install_cell = _build_cell(f"""
code_version = dbutils.widgets.get("code_version")

PRIVATE_PYPI_USERNAME = dbutils.secrets.get(scope="airml-etl", key="PRIVATE_PYPI_USERNAME")
PRIVATE_PYPI_PASSWORD = dbutils.secrets.get(scope="airml-etl", key="PRIVATE_PYPI_PASSWORD")

%pip install --index-url=https://$PRIVATE_PYPI_USERNAME:$PRIVATE_PYPI_PASSWORD@pypi.curia.ai/ {package}==$code_version
    """.strip())
    return install_cell


def _create_databricks_notebook(
        task: DatabricksTaskDefinition,
        env: str,
        db_client: WorkspaceClient,
        databricks_task_name: str):
    """
    Builds a databricks notebook for a task using the task definition and deploys that notebook to the databricks
    workspace.

    :param task: The task definition
    :param env: The environment to deploy to
    :param db_client: The databricks workspace client
    :param databricks_task_name: The name of the task
    """
    notebook = _build_task_head_notebook(env, task, os.environ["CURIA_API_ENDPOINT"])
    try:
        db_client.workspace.mkdirs(f"/etl/{env}")
    except Exception:
        # already exists
        pass
    notebook_path = f"/etl/{env}/{databricks_task_name}"
    db_client.workspace.import_(
        path=notebook_path,
        format=ExportFormat.JUPYTER,
        content=base64.b64encode(json.dumps(notebook).encode("utf-8")).decode("utf-8"),
        overwrite=True
    )
    return notebook_path



import dataclasses
import datetime
import functools
import inspect
import warnings
from collections import defaultdict
from types import MethodType
from typing import List, Dict, Callable, Tuple, Union
from unittest.mock import MagicMock

from curia.api.swagger_client import PlatformApi, Dataset
import curia.api.swagger_client.models as models_package
import re
from pymaybe import maybe
import uuid
from inspect import getframeinfo, stack

from curia.mock.server import Exception404
# pylint: disable=W0622,W0613
from curia.utils.string import to_snake_case, to_camel_case


def debug(message):
    caller = getframeinfo(stack()[1][0])
    print(f"[MockApi DEBUG: {caller.filename}:{caller.lineno}]: {message}")


class _MockApiInstanceBase:
    def __init__(self):
        seed_data, method_overrides, executables_configuration = MockApiConfiguration.get_configuration()
        self.db = defaultdict(dict)
        self._load_seed_data(seed_data)
        self.method_overrides = method_overrides
        self.executables_configuration = executables_configuration
        self.currently_executing = {}

    def _load_seed_data(self, seed_data):
        """
        Loads a list of seed models into the database

        :param seed_data: List of API models to load
        """
        for model in seed_data:
            self._load_model(model)

    def _get_executable_configuration(self, model_type: type):
        for executable_configuration in self.executables_configuration:
            if executable_configuration.executable_class_name == model_type.__name__:
                return executable_configuration
        return None

    def _is_executable(self, model_type: type):
        return self._get_executable_configuration(model_type) is not None

    def _start_mock_execution(self, id_, model_type):
        executable_configuration = self._get_executable_configuration(model_type)
        self.db[model_type][id_].status = executable_configuration.run_status
        self.currently_executing[(model_type, id_)] = datetime.datetime.now()

    def _update_mock_execution_status(self, id_, model_type):
        executable_configuration = self._get_executable_configuration(model_type)
        if (model_type, id_) not in self.currently_executing:
            return
        current_time = datetime.datetime.now()
        start_time = self.currently_executing[(model_type, id_)]
        elapsed_time = current_time - start_time
        print(f"Elapsed time: {elapsed_time.total_seconds()}")
        if elapsed_time.total_seconds() > executable_configuration.time_to_finish:
            self.currently_executing.pop((model_type, id_))
            if id_ in executable_configuration.ids_to_fail:
                self.db[model_type][id_].status = executable_configuration.failed_status
            else:
                self.db[model_type][id_].status = executable_configuration.complete_status

    def _stop_mock_execution(self, id_, model_type):
        executable_configuration = self._get_executable_configuration(model_type)
        if (model_type, id_) not in self.currently_executing:
            return

        self.currently_executing.pop((model_type, id_))
        self.db[model_type][id_].status = executable_configuration.aborted_status

    def _load_model(self, model):
        self.db[type(model)][model.id] = model
        model_snake_case = to_snake_case(type(model).__name__)
        for key in model.attribute_map.keys():
            value = getattr(model, key)

            if type(value) in MODEL_MAP.values() and f"{key}_id" in model.attribute_map and "id" in value.attribute_map:
                # joined model case
                setattr(model, f"{key}_id", value.id)
                if model_snake_case in value.attribute_map:
                    setattr(value, model_snake_case, model)
                elif f"{model_snake_case}s" in value.attribute_map:
                    if isinstance(getattr(value, f"{model_snake_case}s"), list):
                        getattr(value, f"{model_snake_case}s").append(model)
                    else:
                        setattr(value, f"{model_snake_case}s", [model])
                self._load_model(value)
            if isinstance(value, list) and len(value) > 0 \
                    and hasattr(value[0], 'attribute_map') \
                    and f"{model_snake_case}_id" in value[0].attribute_map:
                # joined model case
                for joined_model in value:
                    setattr(joined_model, f"{model_snake_case}_id", model.id)

    def __getattribute__(self, name):
        try:
            method_overrides = object.__getattribute__(self, "method_overrides")
        except AttributeError:
            # handle case where object is still not fully initialized
            method_overrides = {}
        if name in method_overrides:
            return MethodType(method_overrides[name], self)
        return object.__getattribute__(self, name)

    def resolve_joins(self, models, joins: List[str]):
        if len(models) == 0:
            # no models to join other models to
            return
        for join in joins:
            current_models = models
            current_model_type = type(current_models[0])
            segmentation = join.split(".")
            for segment in segmentation:
                inverted_attribute_map = {value: key for key, value in current_model_type.attribute_map.items()}

                if segment not in inverted_attribute_map:
                    raise AttributeError(f"Cannot resolve join: {join}, {current_model_type} does not have an "
                                         f"attribute corresponding to {segment} (looking for '{segment}')")
                attribute_name = inverted_attribute_map[segment]
                attribute_swagger_type: str = current_model_type.swagger_types[attribute_name]
                if attribute_swagger_type.startswith("list["):
                    attribute_swagger_type = attribute_swagger_type[len("list["):-1]
                    next_model_type = MODEL_MAP[attribute_swagger_type]
                    current_model_type_snake_case = to_snake_case(current_model_type.__name__)
                    if f"{current_model_type_snake_case}_id" not in next_model_type.attribute_map:
                        raise ValueError(f"Cannot resolve join: {join}, could not resolve relationship between "
                                         f"{current_model_type} and type {next_model_type}. It's possible that this is a "
                                         f"many to many relationship, which is not yet supported by this mocking "
                                         f"framework, or it is possible that the relationship does not exist. In order "
                                         f"to resolve the relationship, {next_model_type} must have the attribute "
                                         f"{current_model_type_snake_case}_id so that the reverse relationship can be "
                                         f"established.")
                    next_models = []
                    for individual_model in current_models:
                        associated_models = [
                            value for value in self.db[next_model_type].values()
                            if getattr(value, f"{current_model_type_snake_case}_id") == individual_model.id
                        ]
                        setattr(individual_model, attribute_name, associated_models)
                        next_models.extend(associated_models)
                else:
                    next_model_type = MODEL_MAP[attribute_swagger_type]
                    next_models = []
                    for individual_model in current_models:
                        next_model_id = getattr(individual_model, f"{attribute_name}_id")
                        if next_model_id is None:
                            raise ValueError(f"{attribute_name}_id value on {current_models} is null!")
                        # get next model and set attribute of current model
                        next_model = self.db[next_model_type][next_model_id]
                        setattr(individual_model, attribute_name, next_model)
                        next_models.append(next_model)
                current_models = next_models
                current_model_type = next_model_type


def configure_mock_api(seed_data: List[object], method_overrides: Dict[str, Callable]):
    mock_api_configuration = MockApiConfiguration(seed_data=seed_data, method_overrides=method_overrides)

    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            with mock_api_configuration:
                return fn(*args, **kwargs)

        return wrapper

    return decorator


@dataclasses.dataclass
class ExecutableConfiguration:
    executable_class_name: str
    time_to_finish: float = 3.0
    ids_to_fail: List[str] = dataclasses.field(default_factory=list)
    run_status: str = "RUNNING"
    failed_status: str = "FAILED"
    complete_status: str = "COMPLETE"
    aborted_status: str = "ABORTED"


DEFAULT_EXECUTABLES_CONFIGURATION = [
    ExecutableConfiguration(executable_class_name="ModelJob")
]


class MockApiConfiguration:
    STACK: List['MockApiConfiguration'] = []

    def __init__(self,
                 seed_data: List[object],
                 method_overrides: Dict[str, Callable],
                 executables_configuration: List[ExecutableConfiguration] = None):
        """
        Initialize and __enter__ a MockApiConfiguration to seed the mock API with certain values. MockApiConfiguration
        contexts stack -- e.g.:

        with MockApiConfiguration(..., ...):
            with MockApiConfiguration(..., ...):
                api_instance = MockApiInstance()

        Will have the api_instance configured with the seed data from both configurations and the methods from both
        configurations, with the innermost context having precedence.

        :param seed_data: List of API models to seed into the mock database
        :param method_overrides: Dict mapping strings (the names of methods) to callables that implement those methods.
            The callables should expect to receive self as the first argument
        """
        self.seed_data = seed_data
        self.method_overrides = method_overrides
        if executables_configuration is None:
            executables_configuration = DEFAULT_EXECUTABLES_CONFIGURATION
        self.executables_configuration = executables_configuration

    def __enter__(self):
        MockApiConfiguration.STACK.append(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        MockApiConfiguration.STACK.pop()

    @classmethod
    def get_configuration(cls) -> Tuple[List[object], Dict[str, Callable], List[ExecutableConfiguration]]:
        """
        Retrieves the current seed data and method overrides to instantiate a mock API instance

        :return: seed_data as concatenated seed_datas from all configs in STACK, method_overrides as combined
            dictionary of all method_overrides in STACK, with innermost configs taking precedence
        """
        seed_data = []
        method_overrides = {}
        executables_configuration = []
        for configuration in cls.STACK:
            seed_data.extend(configuration.seed_data)
            method_overrides.update(configuration.method_overrides)
            executables_configuration.extend(configuration.executables_configuration)
        if len(executables_configuration) == 0:
            executables_configuration = DEFAULT_EXECUTABLES_CONFIGURATION
        return seed_data, method_overrides, executables_configuration


def get_models():
    """
    Gets all model classes in autogenerated API by stepping through the package

    :return: List of all model classes in autogenerated API
    """
    return {
        model_name: getattr(models_package, model_name)
        for model_name in dir(models_package)
        if isinstance(getattr(models_package, model_name), type)
    }


MODEL_MAP = get_models()


def fn_get_one(models, model_type):
    """
    Constructs a method for getting a single object from the API mock database for an arbitrary model_type
    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that gets a single model from the API mock database
    """

    def get_one(self, id, **kwargs):
        if id in self.db[model_type]:
            if kwargs.get("join", None) is not None:
                self.resolve_joins([self.db[model_type][id]], kwargs["join"])
            if self._is_executable(model_type):
                self._update_mock_execution_status(id, model_type)
            return self.db[model_type][id]
        raise Exception404("404 Model Not Found")

    return get_one


def fn_get_many(models, model_type):
    """
    Constructs a method for getting many objects from the API mock database for an arbitrary model_type
    See docs to implement more filter functionality:
    https://github.com/nestjsx/crud/wiki/Requests#filter

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that gets several models from the API mock database
    """

    def get_many(self, **kwargs):
        to_return = []
        for model in self.db[model_type].values():
            if kwargs.get("filter", None) is not None:
                filter = kwargs["filter"]
                for f in filter:
                    field, condition, value = f.split("||")
                    field_snake_case = to_snake_case(field)
                    try:
                        model_value = getattr(model, field_snake_case)
                    except AttributeError as e:
                        raise AttributeError(f"Model {model} does not have attribute {field_snake_case}!") from e
                    model_value_type = type(model_value)
                    # attempt to coerce type
                    try:
                        model_value = model_value_type(model_value)
                    except Exception as e:
                        debug(f"Failed to coerce type while processing filters; {e}. ")
                    valid = evaluate_condition(condition, model_value, value)
                    if not valid:
                        break
                else:
                    # for else: https://book.pythontips.com/en/latest/for_-_else.html
                    # only runs if there were no breaks (so no filters were invalid)
                    to_return.append(model)
        if "join" in kwargs:
            self.resolve_joins(to_return, kwargs["join"])
        if self._is_executable(model_type):
            for entity in to_return:
                self._update_mock_execution_status(entity.id, model_type)
        return_type = models[f"GetMany{model_type.__name__}ResponseDto"]
        return return_type(data=to_return, count=len(to_return), total=len(to_return), page=0, page_count=1)

    return get_many


def evaluate_condition(condition, model_value, value):
    if condition == "$eq":
        valid = model_value == value
    elif condition == "$ne":
        valid = model_value != value
    elif condition == "$gt":
        valid = model_value > value
    elif condition == "$lt":
        valid = model_value < value
    elif condition == "$gte":
        valid = model_value >= value
    elif condition == "lte":
        valid = model_value <= value
    elif condition == "starts":
        valid = model_value.startswith(value)
    elif condition == "ends":
        valid = model_value.endswith(value)
    else:
        raise ValueError(f"Condition {condition} not recognized by the mock API. "
                         f"If you need it, implement it. (ask EK for reference as to how to do so)")
    return valid


def convert_model_to_model_cls(model: Union[dict, object], cls):
    """
    Converts a model, potentially as a dict,  to a model class instance. If the model is an object implementing to_dict,
    it will be converted to a dict first and then to the model class.
    """
    if hasattr(model, "to_dict"):
        model = model.to_dict()
    try:
        obj = cls(**model)
    except TypeError as e:
        warnings.warn(f"Failed to convert model {model} to model class {cls}. Error: {e} \n"
                      f"Returning model as a dict instead.")
        return model
    # The object is now a model class instance and has swagger_types, so we can use those to recursively search
    # for nested models and convert them to model class instances
    for attr, attr_type in obj.swagger_types.items():
        if attr_type in MODEL_MAP and getattr(obj, attr) is not None:
            # Recursively convert nested models
            setattr(obj, attr, convert_model_to_model_cls(getattr(obj, attr), MODEL_MAP[attr_type]))
        # handle the case when attr_type represents a list of models
        elif attr_type.startswith("list[") and attr_type[5:-1] in MODEL_MAP and getattr(obj, attr) is not None:
            # Recursively convert nested models
            setattr(
                obj,
                attr,
                [convert_model_to_model_cls(item, MODEL_MAP[attr_type[5:-1]]) for item in getattr(obj, attr)]
            )
    return obj


def fn_create_one(models, model_type):
    """
    Constructs a method for creating one object the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that creates a object in the API mock database
    """

    def create_one(self, body, **kwargs):
        body = convert_model_to_model_cls(body, model_type)
        body.id = str(uuid.uuid4())
        self.db[model_type][body.id] = body
        return body

    return create_one


def fn_create_many(models, model_type):
    """
    Constructs a method for creating many objects in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that creates many objects in the API mock database
    """

    def create_many(self, body, **kwargs):
        if hasattr(body, "bulk"):
            bulk = body.bulk
        else:
            bulk = body["bulk"]
        models = []
        for model in bulk:
            model = convert_model_to_model_cls(model, model_type)
            model.id = str(uuid.uuid4())
            self.db[model_type][model.id] = model
            models.append(model)
        return models

    return create_many


def fn_update_one(models, model_type):
    """
    Constructs a method for updating one object in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that updates one object in the API mock database
    """

    def update_one(self, id, body, **kwargs):
        if hasattr(body, "to_dict"):
            body = body.to_dict()
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        result = self.db[model_type][id]
        for field, value in body.items():
            setattr(result, field, value)
        if self._is_executable(model_type):
            self._update_mock_execution_status(id, model_type)
        return result

    return update_one


def fn_delete_one(models, model_type):
    """
    Constructs a method for deleting one object in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that deletes one object in the API mock database
    """

    def delete_one(self, id, **kwargs):
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        del self.db[model_type][id]
        return {"id": id, "deleted": True}

    return delete_one


def fn_replace_one(models, model_type):
    """
    Constructs a method for replacing one object in the API mock database for an arbitrary model_type

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that replaces one object in the API mock database
    """

    def replace_one(self, body, **kwargs):
        body = convert_model_to_model_cls(body, model_type)
        self.db[model_type][body.id] = body
        if self._is_executable(model_type):
            self._update_mock_execution_status(id, model_type)
        return body

    return replace_one


def fn_start(models, model_type):
    """
    Constructs a method for "starting" one object in the API mock database for an arbitrary model_type. No code
    is actually executed when the object is started, the status is just changed.

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that starts one object in the API mock database
    """
    def start(self: _MockApiInstanceBase, id, **kwargs):
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        if not self._is_executable(model_type):
            raise ValueError("This model type is not executable by default. "
                             f"Please add an ExecutableConfiguration for it ({model_type.__name__}) to the mock API "
                             "instance configuration.")
        self._start_mock_execution(id, model_type)
        return self.db[model_type][id]

    return start


def fn_stop(models, model_type):
    """
    Constructs a method for "stopping" one object in the API mock database for an arbitrary model_type. No code
    is actually executed when the object is stopped, the status is just changed.

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that stops one object in the API mock database
    """

    def stop(self, id, **kwargs):
        if id not in self.db[model_type]:
            raise Exception404("404 Model Not Found")
        if not self._is_executable(model_type):
            raise ValueError("This model type is not executable by default. "
                             f"Please add an ExecutableConfiguration for it ({model_type.__name__}) to the mock API "
                             "instance configuration.")
        self._stop_mock_execution(id, model_type)
        return self.db[model_type][id]

    return stop


def fn_status(models, model_type):
    """
    Constructs a method for getting the status of one object in the API mock database for an arbitrary model_type.

    :param models: Dict mapping api model names to api model classes
    :param model_type: api model class, e.g. ProcessJob
    :return: Method that gets the status of one object in the API mock database
    """

    def status(self, id, **kwargs):
        if id in self.db[model_type]:
            if id not in self.db[model_type]:
                raise Exception404("404 Model Not Found")
            if self._is_executable(model_type):
                self._update_mock_execution_status(id, model_type)
            return self.db[model_type][id]
        raise Exception404("404 Model Not Found")

    return status


def upload_dataset(self, body, **kwargs):
    print("upload_dataset called")
    # TODO: if more advanced mocking functionality is needed for this endpoint, it can be added here
    test_body = Dataset(
        name="Mocked Dataset",
        type="query",
        description="Mocked Dataset",
        organization_id="mocked_organization_id",
    )
    test_body.id = str(uuid.uuid4())
    self.db[Dataset][test_body.id] = test_body
    return test_body


def default_functionality_builder(method, models):
    """
    Tries to parse the return type of a method from its docstring and creates a method that returns a MagicMock
    of that return type

    :param method: Method to mock
    :param models: List of models
    :return: function that mocks the default functionality of a model
    """
    return_line = [line.strip() for line in maybe(method.__doc__).split("\n").or_else("") if ":return:" in line]
    return_type = maybe(return_line or None).split()[1].or_else(None)

    if return_type is not None and return_type in models:
        @functools.wraps(method)
        def function(self, *args, **kwargs):
            return MagicMock(models[return_type])
    else:
        if return_type is not None:
            debug(f"{return_type} of method {method} is not a model; reverting to default MagicMock")

        @functools.wraps(method)
        def function(self, *args, **kwargs):
            return MagicMock()
    return function


# matches model names to functionality builders
regex_to_functionality_builder = {
    r"get_one_base_([a-z_]*)_controller_\1": fn_get_one,
    r"get_many_base_([a-z_]*)_controller_\1": fn_get_many,
    r"create_one_base_([a-z_]*)_controller_\1": fn_create_one,
    r"create_many_base_([a-z_]*)_controller_\1": fn_create_many,
    r"update_one_base_([a-z_]*)_controller_\1": fn_update_one,
    r"delete_one_base_([a-z_]*)_controller_\1": fn_delete_one,
    r"replace_one_base_([a-z_]*)_controller_\1": fn_replace_one,
    r"([a-z_]*)_controller_start": fn_start,
    r"([a-z_]*)_controller_stop": fn_stop,
    r"([a-z_]*)_controller_status": fn_status,
    r"dataset_controller_upload": upload_dataset,
}


def get_functionality(models, method_name, method):
    """
    Parses a method name and builds the appropriate functionality around it, e.g. methods that look like
    create_one_base_{}_controller_{} will actually create a {} in the mock database

    :param models: List of parsed models
    :param method_name: Name of method to implement
    :param method: Actual method to implement
    :return: boolean representing whether functionality was parsed and created, Method mocking functionality
    """
    for pattern, functionality_builder in regex_to_functionality_builder.items():
        compiled_pattern = re.compile(pattern)
        match = compiled_pattern.fullmatch(method_name)
        if match:
            if len(match.groups()) == 0:
                # if there are no groups, this method is hardcoded for a specific model
                # e.g. dataset_controller_upload
                return True, functionality_builder
            model_type_snake_case = match.group(1)
            model_type_camel_case = to_camel_case(model_type_snake_case)
            if model_type_camel_case not in models:
                print(f"Could not find model for method {method_name}! "
                      f"(parsed snake case: {model_type_snake_case})"
                      f"(parsed camel case: {model_type_camel_case})")
                break
            return True, functionality_builder(models, models[model_type_camel_case])
    return False, default_functionality_builder(method, models)


def build_method_dict():
    """
    Builds a dictionary of methods on the API and mocked functionality for each method
    :return: Dictionary of methods that can be used to mock the API
    """
    method_dict = {}
    debug(f"Found {len(MODEL_MAP.keys())} models")
    n_specialized = 0
    for method_name, method in inspect.getmembers(PlatformApi, predicate=inspect.isfunction):
        if method_name.startswith("__"):
            continue
        specialized, new_method_functionality = get_functionality(MODEL_MAP, method_name, method)
        method_dict[method_name] = new_method_functionality
        if specialized:
            n_specialized += 1

    debug(f"Found {len(method_dict.keys())} methods, {n_specialized} of which have autogenerated functionality.")
    return method_dict


def get_mock_api_instance_class():
    """
    Builds the mock api instance class by reading the actual api instance
    :return: MockApiInstance class
    """
    method_dict = build_method_dict()

    return type("MockApiInstance", (_MockApiInstanceBase,), method_dict)


MockApiInstance = get_mock_api_instance_class()



import functools
import os
import secrets
import subprocess
import time
from pathlib import Path
from typing import Callable, Any, Optional


TEMP_DOCKER_REGISTRY_SUBFOLDER = "docker_registry"

DOCKER_REGISTRY_VERSION = "2.8.2"


def _generate_random_password():
    """
    Generates a random password for the docker container. NOT SECURE.
    """
    return secrets.token_urlsafe(20)


def _setup_local_docker_registry_credentials(
        mock_registry_folder: Path,
        username: str = 'local_test',
        password: Optional[str] = None):
    """
    Sets up the local docker registry credentials.
    :param mock_registry_folder: The temp folder to use for the mock registry.
    :param username: The username to use for the local docker registry.
    :param password: The password to use for the local docker registry. If None, a random password will be generated.
    """
    # create the auth directory
    auth_folder = mock_registry_folder / "auth"
    auth_folder.mkdir(parents=True, exist_ok=True)
    # generate the password if necessary
    if password is None:
        password = _generate_random_password()
    # docker run --entrypoint htpasswd registry:2 -Bbn local_test password > auth/htpasswd
    try:
        result = subprocess.run(
            [
                "docker",
                "run",
                "--entrypoint",
                "htpasswd",
                "httpd:2",
                "-Bbn",
                username,
                password
            ],
            capture_output=True,
            check=True
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to generate htpasswd file: {e}\n{e.stderr}") from e
    # write the password to the file
    with open(auth_folder / "htpasswd", "w", encoding="utf-8") as f:
        f.write(result.stdout.decode("utf-8"))
    return username, password, auth_folder


def setup_local_docker_registry(
        port: int = 3001,
        temp_folder: Path = Path("./temp/"),
        username: str = 'local_test',
        password: Optional[str] = None
):
    """
    Sets up a local docker registry for testing purposes running on the given port.
    :param port: The port to run the registry on.
    :param temp_folder: The temp folder to use for the mock registry.
    :param username: The username to use for the local docker registry.
    :param password: The password to use for the local docker registry. If None, a random password will be generated.

    :return: tuple of the container id and a dictionary containing the username, password, and host that the docker
    registry is running on.
    """
    mock_registry_folder = temp_folder / TEMP_DOCKER_REGISTRY_SUBFOLDER
    # create the mock registry folder
    mock_registry_folder.mkdir(parents=True, exist_ok=True)
    username, password, auth_folder = _setup_local_docker_registry_credentials(
        mock_registry_folder=mock_registry_folder,
        username=username,
        password=password
    )
    try:
        process = subprocess.run(
            [
                "docker",
                "run",
                "-d",
                "-p",
                f"{port}:5000",
                "--name",
                "mock-registry",
                "-v",
                f"{auth_folder.resolve()}:/auth",
                "-e",
                "REGISTRY_AUTH=htpasswd",
                "-e",
                "REGISTRY_AUTH_HTPASSWD_REALM=Registry Realm",
                "-e",
                "REGISTRY_AUTH_HTPASSWD_PATH=/auth/htpasswd",
                f"registry:{DOCKER_REGISTRY_VERSION}"
            ],
            capture_output=True,
            check=True
        )
        # wait a second for the container to start
        time.sleep(1)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to start docker registry: {e}\n{e.stderr}") from e
    # get the container id from the process
    container_id = process.stdout.decode("utf-8").strip()
    return container_id, {
        "username": username,
        "password": password,
        "host": f"localhost:{port}"
    }


def teardown_local_docker_registry(
        container_id: str,
        temp_folder: Path = Path("./temp/")
):
    """
    Tears down the local docker registry running on the given port and deletes the relevant temp files.
    """
    registry_folder = temp_folder / TEMP_DOCKER_REGISTRY_SUBFOLDER

    #  docker stop mock-registry
    try:
        subprocess.run(
            [
                "docker",
                "stop",
                container_id
            ],
            capture_output=True,
            check=True
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to stop docker registry: {e}\n{e.stderr}") from e
    # docker rm mock-registry
    try:
        subprocess.run(
            [
                "docker",
                "rm",
                container_id
            ],
            capture_output=True,
            check=True
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to remove docker registry: {e}\n{e.stderr}") from e
    # rm -rf registry
    os.system(f"rm -rf {registry_folder}")


def using_local_docker_registry(port: int = 3001,
                                temp_folder: Path = Path("./temp/"),
                                username: str = 'local_test',
                                password: Optional[str] = None):
    """
    Decorator for tests that require a local docker registry. Automatically sets up the registry, runs the wrapped
    function, and tears down the registry.

    :param port: The port to run the registry on.
    :param temp_folder: The temp folder to use for the mock registry.
    :param username: The username to use for the local docker registry.
    :param password: The password to use for the local docker registry. If None, a random password will be generated.
    """
    def decorator(fn) -> Callable:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs) -> Any:
            registry_id, credentials = setup_local_docker_registry(
                port=port,
                temp_folder=temp_folder,
                username=username,
                password=password
            )
            try:
                return fn(*args, **kwargs, docker_registry_credentials=credentials)
            finally:
                teardown_local_docker_registry(
                    registry_id,
                    temp_folder=temp_folder
                )

        return wrapper
    return decorator



import functools
import inspect
from typing import Tuple, List, Dict, Callable, Union, Any
from unittest.mock import MagicMock

from curia.api.swagger_client import PlatformApi, ApiClient
from curia.mock.api import build_method_dict, MockApiConfiguration, MockApiInstance
from curia.mock.server import ParseAndExtractDataSummary, Exception404, using_test_server
from curia.utils.string import to_snake_case


def build_mock_arguments(method: callable) -> Dict[str, Any]:
    """
    Builds a dictionary of mock arguments for a method, using the "inspect" module to parse the method signature
    There is no guarantee that this will work for all methods, but it should work for the ones we use in the
    PlatformApi

    :param method: Method to build mock arguments for
    :return: Dictionary of mock arguments. If the method has a default value for an argument, the default value is used.
    Otherwise, a MagicMock is used.
    """
    mock_arguments = {}
    for arg_name, arg in inspect.signature(method).parameters.items():
        if arg_name == "self":
            continue
        if arg.kind == inspect.Parameter.VAR_KEYWORD:
            continue
        if arg.kind == inspect.Parameter.VAR_POSITIONAL:
            continue
        if arg.default is not inspect.Parameter.empty:
            mock_arguments[arg_name] = arg.default
        else:
            mock_arguments[arg_name] = MagicMock()
    return mock_arguments


def match_call_pattern_to_method_name(
        call_pattern: Tuple[str, str],
        call_pattern_to_method_name: Dict[Tuple, str]) \
        -> Tuple[str, dict]:
    """
    Matches a URL call pattern to a method name, e.g. (/models/some-id-here, GET) -> get_one_base_model_controller_model
    :param call_pattern: Realized call pattern to match, e.g. (/models/some-id-here, GET)
    :param call_pattern_to_method_name: Dictionary of unrealized call patterns (e.g. /models/{id}) to method names
    :return: Method name that matches the call pattern
    """
    url, call_request_method = call_pattern
    call_pattern_components = url.split("/")[1:]  # remove leading slash
    for pattern, method_name in call_pattern_to_method_name.items():
        url_pattern, request_method = pattern
        if request_method != call_request_method:
            continue
        url_pattern_components = url_pattern.split("/")[1:]  # remove leading slash
        if len(url_pattern_components) != len(call_pattern_components):
            continue
        kwargs = {}
        for pattern_component, call_pattern_component in zip(url_pattern_components, call_pattern_components):
            if pattern_component.startswith("{") and pattern_component.endswith("}"):
                kwargs[pattern_component[1:-1]] = call_pattern_component
                continue
            if pattern_component != call_pattern_component:
                break
        else:
            return method_name, kwargs
    raise ValueError(f"Could not match call pattern {call_pattern} to any method!")


class Struct:
    """
    A simple class that allows for attribute access to a dictionary
    """

    def __init__(self, **entries):
        self._entries = entries

    def to_dict(self) -> dict:
        return self._entries

    def __setattr__(self, key, value) -> None:
        if key != "_entries":
            self._entries[key] = value
        super().__setattr__(key, value)

    def __getattr__(self, item) -> Any:
        if item == "_entries":
            return super().__getattribute__(item)
        if item in self._entries:
            return self._entries[item]
        return super().__getattribute__(item)

    def __repr__(self) -> str:
        return str(self._entries)


class MockResponseMapper:
    """
    Builds a function capable of handling responses from the mock server in mock.server.py by hijacking
    the actual api instance and checking which requests are made to the server.
    """
    _mock_response_mapper = None

    def __init__(self):
        hijacked_api_instance = self.get_api_client_hijacker()

        self.method_dict = build_method_dict()

        self.call_pattern_to_method_name = build_call_pattern_dict(hijacked_api_instance, self.method_dict)
        self.mock_api_instance = MockApiInstance()

    def get_api_client_hijacker(self) -> PlatformApi:
        """
        Returns an instance of the api client that has been hijacked to record the last call made to the server
        :return: Hijacked api instance with a hijacked api client
        """
        hijacked_api_instance = PlatformApi()

        class ApiClientHijacker(ApiClient):
            """
            Hijacks the api client to record the last call made to the server
            """

            def __init__(self):
                self.last_call = None
                super().__init__()

            def call_api(self,
                         resource_path, method,
                         path_params=None, query_params=None, header_params=None,
                         body=None, post_params=None, files=None,
                         response_type=None, auth_settings=None, async_req=None,
                         _return_http_data_only=None, collection_formats=None,
                         _preload_content=True, _request_timeout=None) -> (object, int):
                self.last_call = (resource_path, method)
                return None, 200

        hijacked_api_instance.api_client = ApiClientHijacker()
        return hijacked_api_instance

    def response_map(self, request: ParseAndExtractDataSummary) -> Union[dict, list]:
        """
        Performs the actual mapping of the request to a response. This is the function that is called by the mock server
        to determine the response to a request. We first match the request to a method name, then we call the method
        on the mock api instance and return the result, which is then serialized to JSON and returned to the client.

        The request's body is prepped to be passed to the mock api instance by converting it to an object with
        attributes named using snake case. This is done because the mock api instance expects the request body to be
        an object with attributes named using snake case, but the client sends the request body as a dictionary with
        attributes named using lower camel case.

        The result of the mock api instance method is also prepped to be returned to the client by converting it to a
        dictionary with attributes named using lower camel case. This is done because the client expects the response
        body to be JSON with attributes named using lower camel case, but the mock api instance returns the response
        body as an object with attributes named using snake case.

        :param request: The request to map to a response
        """
        try:
            method_name, kwargs = match_call_pattern_to_method_name(
                (request.path, request.method),
                self.call_pattern_to_method_name
            )
        except ValueError as e:
            raise Exception404(f"Could not match call pattern {request.method} {request.path} to any method!") from e
        if not hasattr(self.mock_api_instance, method_name):
            raise Exception(f"Method {method_name} not found in mock api instance!")
        body_dict = {}
        if request.content_qs:
            if "bulk" in request.content_qs:
                body_dict = {
                    "body": {
                        "bulk": prep_for_obj(request.content_qs["bulk"])
                    }
                }
            else:
                body_dict = {"body": prep_for_obj(request.content_qs)}
        res = getattr(self.mock_api_instance, method_name)(
            **kwargs,
            **request.url_qs,
            **body_dict
        )
        return prep_for_json(res)

    @classmethod
    def get(cls) -> "MockResponseMapper":
        """
        Singleton getter

        :return: MockResponseMapper instance. Note that this is a singleton, so the same instance will be returned
        every time. This is necessary because the response_map callable is stateless, but we don't want to set up the
        same mock api instance and call pattern mapping every time we need to use it.
        """
        if cls._mock_response_mapper is None:
            cls._mock_response_mapper = cls()
        return cls._mock_response_mapper


def prep_for_obj(obj: Any) -> Any:
    """
    Prepares a dict for use as an object by converting all keys to snake case. This is necessary because the
    SDK models use snake case, but the actual api uses lower camel case.
    The outputs of this function are used to create SDK models, so they must match the format of the SDK models.

    :param obj: Object to prepare. Can be a dict, list, or a single value (str, int, etc.)
    :return: object prepared for use as an SDK model
    """
    if isinstance(obj, dict):
        return {to_snake_case(key): prep_for_obj(value) for key, value in obj.items()}
    if isinstance(obj, list):
        return [prep_for_obj(item) for item in obj]
    return obj


def prep_for_json(obj: Any) -> Any:
    """
    Prepares a model class for json serialization and return to the client by:
    - Converting objects to dicts using their attributes listed in the attribute_map
    - Converting the attribute names according to the attribute_map of the object
    - Recursively applying the above to all values in a dict, list, or sub-object

    This is necessary because the SDK models generally snake case, but the actual api generally uses lower camel case.
    The outputs of this function are returned to the client as json, so they must be in the same format as the actual
    api. We use the attribute_map of the SDK models to convert the attribute names to their API equivalents.

    The objects also have a to_dict method, but this method is overzealous and recursively converts all sub-objects to
    dicts, which is not desirable because then we can't tell how to convert the attribute names of the sub-objects
    as they lose the attribute_map of their class.

    Instead, we rely on this function to eventually convert the sub-objects to dicts, but only when it is safe to do so
    and the attribute names can be converted correctly.

    :param obj: Object to prepare. Can be a dict, list, object with attribute_map, or a single value (str, int, etc.)
    :return: object prepared for json serialization
    """
    if isinstance(obj, list):
        return [prep_for_json(item) for item in obj]
    if isinstance(obj, dict):
        return {key: prep_for_json(value) for key, value in obj.items()}
    if hasattr(obj, "attribute_map"):
        return {
            api_attr_name: prep_for_json(getattr(obj, sdk_attr_name))
            for sdk_attr_name, api_attr_name in obj.attribute_map.items()
        }
    return obj


def mock_api_request_handler_response_map(
        request: ParseAndExtractDataSummary,
        metadata: MockApiConfiguration) -> Union[dict, list]:
    """
    This function is compatible with the response_map parameter of the TestServer class in mock.server.py. When using
    the TestServer class, the metadata parameter should be a MockApiConfiguration instance specifying the initial
    state of the mock api instance which will be utilized by the MockApiInstance class under the hood.
    :param request: The request to handle. Will be automatically passed by the TestServer class.
    :param metadata: The MockApiConfiguration instance specifying the initial state of the MockApiInstance utilized
    under the hood. The value of this parameter will be determined by the response_map_metadata parameter of the
    TestServer class, or by the response_map_metadata parameter of the configure_mock_api_server function.

    :return: The response to return to the client. Should be a dict or list, turnable into json.
    """
    with metadata:
        return MockResponseMapper.get().response_map(request)


def build_call_pattern_dict(hijacked_api_instance: PlatformApi, methods: list) -> Dict[Tuple[str, str], str]:
    """
    Builds a dictionary that maps a call pattern (resource path, method) to a method name in the PlatformApi class
    and thereby to a method in the MockApiInstance class. This is done by hijacking the api client of the PlatformApi
    instance and checking which requests are made to the server when a method is called. The call pattern is then
    mapped to the method name of the method that was called.

    In order to call the methods of the PlatformApi instance, mock arguments are built for each method. These mock
    arguments are built by inspecting the method signature and building mock arguments for each parameter.
    :param hijacked_api_instance: The PlatformApi instance. The api client of this instance should already be hijacked
    with an ApiClientHijacker
    :param methods: A list of method names in the PlatformApi class that should be called in order to build the call
    pattern dictionary. Method names that end with "_with_http_info" are ignored as these methods are not intended to
    be called directly and are not mocked in the MockApiInstance class, so calling them would result in an error.

    :return: A dictionary mapping a call pattern (resource path, method) to a method name in the PlatformApi class
    """
    call_pattern_to_method_name = {}
    for method_name in methods:
        if method_name.endswith("_with_http_info"):
            continue
        method = getattr(hijacked_api_instance, method_name)
        mock_arguments = build_mock_arguments(method)
        method(**mock_arguments)
        if method_name.endswith("_0"):
            # some unusual duplicates exist ending with _0 that gunk up the process if present, so we skip them.
            continue
        call_pattern_to_method_name[hijacked_api_instance.api_client.last_call] = method_name
    return call_pattern_to_method_name


def configure_mock_api_server(seed_data: List[object],
                              method_overrides: Dict[str, Callable],
                              port=3000,
                              log_dir="temp/") -> Callable:
    """
    Decorator for test functions that configures a mock api server for the duration of the test function. The mock
    api server is configured to contain the seed data provided in its database. In a similar manner to MockApiInstance,
    the mock api server can be configured to override the functionality of certain methods by providing a callable
    for the method in the method_overrides dict. The mock api server is configured to run on the port provided and
    to log requests to the log_dir provided.

    Under the hood, this decorator uses the MockApiInstance class to simulate the functionality of the api server.
    So when a method on a client's real sdk is called, the client's sdk will translate the method call into a request
    to the mock api server. The mock api server will then translate the request *back* into its original method call
    and execute the method call on the MockApiInstance instance. The MockApiInstance instance will then execute the
    inferred or supplied method override, if any, and return the result to the mock server, which will then convert
    the result (an object) into the appropriate json response and return it to the client's sdk, which will then
    convert the json response into an object and return it to the client.

    This decorator combines the functionality of the MockApiInstance class and using_test_server decorator.

    The using_test_server decorator can be used to set up an arbitrary mock server for testing purposes. However, it
    requires the user to provide a response_map callable that maps requests to responses. This decorator provides
    a default response_map callable that maps requests to responses based on the seed data and method overrides
    provided, using the MockApiInstance class.

    :param seed_data: List of objects to seed the mock api server with. The objects should be instances of
    the classes in the curia.api.swagger_client.models package.
    :param method_overrides: Dict of method names to override with custom functions. The method names should be
    the same as the method names in the PlatformApi class. The custom functions should take the same arguments as
    the original methods and return the same type as the original methods. The custom functions should not call
    the original methods.
    :param port: Port to run the mock api server on
    :param log_dir: Directory to log requests to
    """
    mock_api_configuration = MockApiConfiguration(seed_data=seed_data, method_overrides=method_overrides)

    def decorator(fn) -> Callable:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs) -> Any:
            with mock_api_configuration:
                fn_with_test_server = using_test_server(
                    port=port,
                    log_dir=log_dir,
                    response_map=mock_api_request_handler_response_map,
                    response_map_metadata=mock_api_configuration
                )(fn)
                return fn_with_test_server(*args, **kwargs)

        return wrapper

    return decorator



# coding: utf-8

"""
    Curia Platform API

    These are the docs for the curia platform API. To test, generate an authorization token first.  # noqa: E501

    OpenAPI spec version: 3.8.1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from __future__ import absolute_import

import unittest

import curia.api.curia.api.curia.api.swagger_client
from curia.api.curia.api.curia.api.swagger_client.api.tasks_api import TasksApi  # noqa: E501
from curia.api.curia.api.curia.api.swagger_client.rest import ApiException


class TestTasksApi(unittest.TestCase):
    """TasksApi unit test stubs"""

    def setUp(self):
        self.api = TasksApi()  # noqa: E501

    def tearDown(self):
        pass

    def test_create_many_base_task_controller_task(self):
        """Test case for create_many_base_task_controller_task

        Create multiple Tasks  # noqa: E501
        """
        pass

    def test_create_one_base_task_controller_task(self):
        """Test case for create_one_base_task_controller_task

        Create a single Task  # noqa: E501
        """
        pass

    def test_delete_one_base_task_controller_task(self):
        """Test case for delete_one_base_task_controller_task

        Delete a single Task  # noqa: E501
        """
        pass

    def test_get_many_base_task_controller_task(self):
        """Test case for get_many_base_task_controller_task

        Retrieve multiple Tasks  # noqa: E501
        """
        pass

    def test_get_one_base_task_controller_task(self):
        """Test case for get_one_base_task_controller_task

        Retrieve a single Task  # noqa: E501
        """
        pass

    def test_replace_one_base_task_controller_task(self):
        """Test case for replace_one_base_task_controller_task

        Replace a single Task  # noqa: E501
        """
        pass

    def test_task_controller_create_task_from_databricks_job(self):
        """Test case for task_controller_create_task_from_databricks_job

        Create task from Databricks job  # noqa: E501
        """
        pass

    def test_task_controller_get_awsecr_repos(self):
        """Test case for task_controller_get_awsecr_repos

        List ECR repos  # noqa: E501
        """
        pass

    def test_task_controller_list_databricks_jobs(self):
        """Test case for task_controller_list_databricks_jobs

        List Databricks jobs  # noqa: E501
        """
        pass

    def test_update_one_base_task_controller_task(self):
        """Test case for update_one_base_task_controller_task

        Update a single Task  # noqa: E501
        """
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    Curia Platform API

    These are the docs for the curia platform API. To test, generate an authorization token first.  # noqa: E501

    OpenAPI spec version: 3.10.4
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class Population(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'id': 'str',
        'last_updated_by': 'str',
        'created_by': 'str',
        'created_at': 'datetime',
        'updated_at': 'datetime',
        'archived_at': 'datetime',
        'version': 'float',
        'population_store_id': 'str',
        'primary_key': 'str',
        'partion_key': 'list[str]',
        'population_store': 'PopulationStore',
        'model_populations': 'list[ModelPopulation]',
        'outcomes': 'list[OutcomeDefinition]',
        'interventions': 'list[InterventionDefinition]',
        'cohorts': 'list[CohortDefinition]',
        'organization_id': 'str',
        'organization': 'Organization'
    }

    attribute_map = {
        'id': 'id',
        'last_updated_by': 'lastUpdatedBy',
        'created_by': 'createdBy',
        'created_at': 'createdAt',
        'updated_at': 'updatedAt',
        'archived_at': 'archivedAt',
        'version': 'version',
        'population_store_id': 'populationStoreId',
        'primary_key': 'primaryKey',
        'partion_key': 'partionKey',
        'population_store': 'populationStore',
        'model_populations': 'modelPopulations',
        'outcomes': 'outcomes',
        'interventions': 'interventions',
        'cohorts': 'cohorts',
        'organization_id': 'organizationId',
        'organization': 'organization'
    }

    def __init__(self, id=None, last_updated_by=None, created_by=None, created_at=None, updated_at=None, archived_at=None, version=None, population_store_id=None, primary_key=None, partion_key=None, population_store=None, model_populations=None, outcomes=None, interventions=None, cohorts=None, organization_id=None, organization=None):  # noqa: E501
        """Population - a model defined in Swagger"""  # noqa: E501
        self._id = None
        self._last_updated_by = None
        self._created_by = None
        self._created_at = None
        self._updated_at = None
        self._archived_at = None
        self._version = None
        self._population_store_id = None
        self._primary_key = None
        self._partion_key = None
        self._population_store = None
        self._model_populations = None
        self._outcomes = None
        self._interventions = None
        self._cohorts = None
        self._organization_id = None
        self._organization = None
        self.discriminator = None
        if id is not None:
            self.id = id
        if last_updated_by is not None:
            self.last_updated_by = last_updated_by
        if created_by is not None:
            self.created_by = created_by
        if created_at is not None:
            self.created_at = created_at
        if updated_at is not None:
            self.updated_at = updated_at
        if archived_at is not None:
            self.archived_at = archived_at
        if version is not None:
            self.version = version
        self.population_store_id = population_store_id
        if primary_key is not None:
            self.primary_key = primary_key
        if partion_key is not None:
            self.partion_key = partion_key
        if population_store is not None:
            self.population_store = population_store
        if model_populations is not None:
            self.model_populations = model_populations
        if outcomes is not None:
            self.outcomes = outcomes
        if interventions is not None:
            self.interventions = interventions
        if cohorts is not None:
            self.cohorts = cohorts
        self.organization_id = organization_id
        if organization is not None:
            self.organization = organization

    @property
    def id(self):
        """Gets the id of this Population.  # noqa: E501


        :return: The id of this Population.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Population.


        :param id: The id of this Population.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def last_updated_by(self):
        """Gets the last_updated_by of this Population.  # noqa: E501


        :return: The last_updated_by of this Population.  # noqa: E501
        :rtype: str
        """
        return self._last_updated_by

    @last_updated_by.setter
    def last_updated_by(self, last_updated_by):
        """Sets the last_updated_by of this Population.


        :param last_updated_by: The last_updated_by of this Population.  # noqa: E501
        :type: str
        """

        self._last_updated_by = last_updated_by

    @property
    def created_by(self):
        """Gets the created_by of this Population.  # noqa: E501


        :return: The created_by of this Population.  # noqa: E501
        :rtype: str
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this Population.


        :param created_by: The created_by of this Population.  # noqa: E501
        :type: str
        """

        self._created_by = created_by

    @property
    def created_at(self):
        """Gets the created_at of this Population.  # noqa: E501


        :return: The created_at of this Population.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Population.


        :param created_at: The created_at of this Population.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this Population.  # noqa: E501


        :return: The updated_at of this Population.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this Population.


        :param updated_at: The updated_at of this Population.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    @property
    def archived_at(self):
        """Gets the archived_at of this Population.  # noqa: E501


        :return: The archived_at of this Population.  # noqa: E501
        :rtype: datetime
        """
        return self._archived_at

    @archived_at.setter
    def archived_at(self, archived_at):
        """Sets the archived_at of this Population.


        :param archived_at: The archived_at of this Population.  # noqa: E501
        :type: datetime
        """

        self._archived_at = archived_at

    @property
    def version(self):
        """Gets the version of this Population.  # noqa: E501


        :return: The version of this Population.  # noqa: E501
        :rtype: float
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this Population.


        :param version: The version of this Population.  # noqa: E501
        :type: float
        """

        self._version = version

    @property
    def population_store_id(self):
        """Gets the population_store_id of this Population.  # noqa: E501


        :return: The population_store_id of this Population.  # noqa: E501
        :rtype: str
        """
        return self._population_store_id

    @population_store_id.setter
    def population_store_id(self, population_store_id):
        """Sets the population_store_id of this Population.


        :param population_store_id: The population_store_id of this Population.  # noqa: E501
        :type: str
        """
        if population_store_id is None:
            raise ValueError("Invalid value for `population_store_id`, must not be `None`")  # noqa: E501

        self._population_store_id = population_store_id

    @property
    def primary_key(self):
        """Gets the primary_key of this Population.  # noqa: E501


        :return: The primary_key of this Population.  # noqa: E501
        :rtype: str
        """
        return self._primary_key

    @primary_key.setter
    def primary_key(self, primary_key):
        """Sets the primary_key of this Population.


        :param primary_key: The primary_key of this Population.  # noqa: E501
        :type: str
        """

        self._primary_key = primary_key

    @property
    def partion_key(self):
        """Gets the partion_key of this Population.  # noqa: E501


        :return: The partion_key of this Population.  # noqa: E501
        :rtype: list[str]
        """
        return self._partion_key

    @partion_key.setter
    def partion_key(self, partion_key):
        """Sets the partion_key of this Population.


        :param partion_key: The partion_key of this Population.  # noqa: E501
        :type: list[str]
        """

        self._partion_key = partion_key

    @property
    def population_store(self):
        """Gets the population_store of this Population.  # noqa: E501


        :return: The population_store of this Population.  # noqa: E501
        :rtype: PopulationStore
        """
        return self._population_store

    @population_store.setter
    def population_store(self, population_store):
        """Sets the population_store of this Population.


        :param population_store: The population_store of this Population.  # noqa: E501
        :type: PopulationStore
        """

        self._population_store = population_store

    @property
    def model_populations(self):
        """Gets the model_populations of this Population.  # noqa: E501


        :return: The model_populations of this Population.  # noqa: E501
        :rtype: list[ModelPopulation]
        """
        return self._model_populations

    @model_populations.setter
    def model_populations(self, model_populations):
        """Sets the model_populations of this Population.


        :param model_populations: The model_populations of this Population.  # noqa: E501
        :type: list[ModelPopulation]
        """

        self._model_populations = model_populations

    @property
    def outcomes(self):
        """Gets the outcomes of this Population.  # noqa: E501


        :return: The outcomes of this Population.  # noqa: E501
        :rtype: list[OutcomeDefinition]
        """
        return self._outcomes

    @outcomes.setter
    def outcomes(self, outcomes):
        """Sets the outcomes of this Population.


        :param outcomes: The outcomes of this Population.  # noqa: E501
        :type: list[OutcomeDefinition]
        """

        self._outcomes = outcomes

    @property
    def interventions(self):
        """Gets the interventions of this Population.  # noqa: E501


        :return: The interventions of this Population.  # noqa: E501
        :rtype: list[InterventionDefinition]
        """
        return self._interventions

    @interventions.setter
    def interventions(self, interventions):
        """Sets the interventions of this Population.


        :param interventions: The interventions of this Population.  # noqa: E501
        :type: list[InterventionDefinition]
        """

        self._interventions = interventions

    @property
    def cohorts(self):
        """Gets the cohorts of this Population.  # noqa: E501


        :return: The cohorts of this Population.  # noqa: E501
        :rtype: list[CohortDefinition]
        """
        return self._cohorts

    @cohorts.setter
    def cohorts(self, cohorts):
        """Sets the cohorts of this Population.


        :param cohorts: The cohorts of this Population.  # noqa: E501
        :type: list[CohortDefinition]
        """

        self._cohorts = cohorts

    @property
    def organization_id(self):
        """Gets the organization_id of this Population.  # noqa: E501


        :return: The organization_id of this Population.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this Population.


        :param organization_id: The organization_id of this Population.  # noqa: E501
        :type: str
        """
        if organization_id is None:
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def organization(self):
        """Gets the organization of this Population.  # noqa: E501


        :return: The organization of this Population.  # noqa: E501
        :rtype: Organization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this Population.


        :param organization: The organization of this Population.  # noqa: E501
        :type: Organization
        """

        self._organization = organization

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Population, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Population):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Curia Platform API

    These are the docs for the curia platform API. To test, generate an authorization token first.  # noqa: E501

    OpenAPI spec version: 3.10.4
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class ZipData(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'zip': 'str',
        'patient_count': 'float',
        'outcome_count': 'float',
        'intervention_count': 'float'
    }

    attribute_map = {
        'zip': 'zip',
        'patient_count': 'patientCount',
        'outcome_count': 'outcomeCount',
        'intervention_count': 'interventionCount'
    }

    def __init__(self, zip=None, patient_count=None, outcome_count=None, intervention_count=None):  # noqa: E501
        """ZipData - a model defined in Swagger"""  # noqa: E501
        self._zip = None
        self._patient_count = None
        self._outcome_count = None
        self._intervention_count = None
        self.discriminator = None
        self.zip = zip
        self.patient_count = patient_count
        self.outcome_count = outcome_count
        self.intervention_count = intervention_count

    @property
    def zip(self):
        """Gets the zip of this ZipData.  # noqa: E501


        :return: The zip of this ZipData.  # noqa: E501
        :rtype: str
        """
        return self._zip

    @zip.setter
    def zip(self, zip):
        """Sets the zip of this ZipData.


        :param zip: The zip of this ZipData.  # noqa: E501
        :type: str
        """
        if zip is None:
            raise ValueError("Invalid value for `zip`, must not be `None`")  # noqa: E501

        self._zip = zip

    @property
    def patient_count(self):
        """Gets the patient_count of this ZipData.  # noqa: E501


        :return: The patient_count of this ZipData.  # noqa: E501
        :rtype: float
        """
        return self._patient_count

    @patient_count.setter
    def patient_count(self, patient_count):
        """Sets the patient_count of this ZipData.


        :param patient_count: The patient_count of this ZipData.  # noqa: E501
        :type: float
        """
        if patient_count is None:
            raise ValueError("Invalid value for `patient_count`, must not be `None`")  # noqa: E501

        self._patient_count = patient_count

    @property
    def outcome_count(self):
        """Gets the outcome_count of this ZipData.  # noqa: E501


        :return: The outcome_count of this ZipData.  # noqa: E501
        :rtype: float
        """
        return self._outcome_count

    @outcome_count.setter
    def outcome_count(self, outcome_count):
        """Sets the outcome_count of this ZipData.


        :param outcome_count: The outcome_count of this ZipData.  # noqa: E501
        :type: float
        """
        if outcome_count is None:
            raise ValueError("Invalid value for `outcome_count`, must not be `None`")  # noqa: E501

        self._outcome_count = outcome_count

    @property
    def intervention_count(self):
        """Gets the intervention_count of this ZipData.  # noqa: E501


        :return: The intervention_count of this ZipData.  # noqa: E501
        :rtype: float
        """
        return self._intervention_count

    @intervention_count.setter
    def intervention_count(self, intervention_count):
        """Sets the intervention_count of this ZipData.


        :param intervention_count: The intervention_count of this ZipData.  # noqa: E501
        :type: float
        """
        if intervention_count is None:
            raise ValueError("Invalid value for `intervention_count`, must not be `None`")  # noqa: E501

        self._intervention_count = intervention_count

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ZipData, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ZipData):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
